/*
  ESP8266 firmware to let TI-Innovator Hub control a Tello or Tello EDU drone
  (based on Tello SDK 1.3 and Tello EDU SDK 2.0
  created by Hans-Martin Hilbig (Mar 2021)
  REV 1.0   based on Rev 2.2 of tello_edu.ino
*/
  String vers = "REV 1.1"; //Initial T3 release
 
#include <ESP8266WiFi.h>
#include <WiFiUDP.h>

// Tello WiFi settings
const char* DEFAULT_PW = "";
const char* TELLO_IP = "192.168.10.1";
const int PORT = 8889;
const int PORTr = 8890;
char* DEFAULT_SSID = "TELLO-58FADA";

// WiFi connection status
const int INIT = 0;
const int CONNECTED = 1;
const int WIFIOK = 2;
const int LOST = 3;
int networkState = INIT;

WiFiUDP Udp;  // Tello's control command UDP Port channel
WiFiUDP Udpr; // Tello's bulk status read command UDP Port channel

// character buffers
char packetBuffer[255]; //buffer to hold incoming UDP packet
char rcvBuffer[255];  //command rcv buffer

// timeouts & LED status blink periods
const int Ctimeout = 14000;  //Tello SDK command timeout
const int wifiTimeout = 30000; //Tello WiFi connect timeout
const int Tretry = 1000;     //Tello reconnect trial wait time
const int LEDcommand = 1000;  //LED status: waiting CLI input
const int LEDretry = 500;  //LED status: Tello connection lost
const int LEDUDP = 100;  //LED status: waiting Tello packet
const int ONBOARD_LED = 2;

// Vars
bool flip = false;
bool debug = false;
bool autorefresh = false;  //Periodic refresh of Tello SDK
bool wififresh = true;   //WiFi hasn't been connected before
int Ctag;   // command queue timeout tag
int LEDtag; // LED toggle timeout tag
String rcvStr;
String Tmessage;
String Rmessage;


void setup() {
  // initialize serial port:
  Serial.begin(115200);
  pinMode(ONBOARD_LED, OUTPUT);
  LEDtag = millis();
}

void loop() {
   // refresh Tello link
  if ((networkState==CONNECTED) and (millis()>Ctag+Ctimeout) and (autorefresh))  {
    if (!TelloOnline(1))  {
      networkState = LOST;
      Serial.println("TELLO LOST!");
    }
    else  {
      Serial.println("TELLO REFRESHED");
    }
    Ctag = millis();
  }
  else  {
    if ((networkState==LOST) and (millis()>Ctag+Tretry) and (autorefresh))  {
      Serial.println("RETRY..");
      if (TelloOnline(1))  {
        Serial.println("TELLO RECONNECTED!");
      }
      Ctag = millis();
    }
  }
  
  // CLI processor ***********************************
  if (Serial.available()) {
    rcvStr = Serial.readString();
    if (debug)  {
      Serial.println("COMMAND RECEIVED: "+rcvStr);
    }
  }
  // Network & system commands
  if (rcvStr == "CONNECT")  {
    if (initSequence())  {
      Serial.println("OK");
    }
    else  {
      Serial.println("ERROR");
    }
    cmdClr();
  } 
  else if (rcvStr == "NETSTATE")  {
    Serial.println("NETWORKSTATE: "+String(networkState));
    cmdClr();
  }
else if (rcvStr == "DEBUG")  {
    debug = true;
    Serial.println("OK");
    cmdClr();
  }
  else if (rcvStr == "NODEBUG")  {
    debug = false;
    Serial.println("OK");
    cmdClr();
  }
  else if (rcvStr == "REFRESH")  {
    autorefresh = true;
    Serial.println("OK");
    cmdClr();
  }
  else if (rcvStr == "NOREFRESH")  {
    autorefresh = false;
    Serial.println("OK");
    cmdClr();
  }
  else if (rcvStr == "FLUSH")  {
    Serial.println(listenMessage(300));
    Serial.println("OK");
    cmdClr();
  }
  else if (rcvStr == "OK?")  {
    if (TelloOnline(1))  {
      Serial.println("TELLO ONLINE!");
    }
    else  {
      Serial.println("TELLO OFFLINE!");
    }
    cmdClr();
  }
  else if (rcvStr == "ESP?")  {
    Serial.println("TELLO CLI "+vers);
    cmdClr();
  }
  else if (rcvStr.substring(0,4) == "SSID")  {   //Tello WiFi SSID
    rcvStr = (rcvStr.substring(5,rcvStr.length()));
    if (rcvStr != DEFAULT_SSID)   {
      rcvStr.toCharArray(DEFAULT_SSID,255);
      networkState = INIT;
      if (debug) {
        Serial.println(DEFAULT_SSID);
      }
    }
    Serial.println("OK");
    cmdClr();
  }
/* ********************************************  
 *          Tello SDK 1.3.0.0 commands        *
 **********************************************/
 
  // Control commands
  else if (rcvStr == "TAKEOFF")  {
    sendCommand("takeoff",20000);
    cmdClr();
  }
  else if (rcvStr == "LAND")  {
    sendCommand("land",20000);
    cmdClr();
  }
  else if (rcvStr == "OFF")  {
    sendCommand("emergency",200);
    Serial.println("OK");
    cmdClr();
  }

  // rc commands
  else if (rcvStr == "ON")  {
    sendCommand("rc -100 -100 -100 100",2000);
    sendCommand("rc 0 0 0 0",200);
    Serial.println("OK");
    cmdClr();
  }
  else if (rcvStr == "NEUTRAL")  {
    sendCommand("rc 0 0 0 0",200);
    Serial.println("OK");
    cmdClr();
  }
  else if (rcvStr.substring(0,2) == "RC")  {
    rcvStr.toLowerCase();
    rcvStr.toCharArray(rcvBuffer,255);
    sendCommand(rcvBuffer,200);
    Serial.println("OK");
    cmdClr();
  }
  else if (rcvStr.substring(0,2) == "CW")  {    //1 - 3600 degrees
    rcvStr.toLowerCase();
    rcvStr.toCharArray(rcvBuffer,255);
    sendCommand(rcvBuffer,14000);
    cmdClr();
  }
  else if (rcvStr.substring(0,3) == "CCW")  {   //1 - 3600 degrees
    rcvStr.toLowerCase();
    rcvStr.toCharArray(rcvBuffer,255);
    sendCommand(rcvBuffer,14000);
    cmdClr();
  }
else if (rcvStr.substring(0,7) == "FORWARD")  { //20 - 500 cm
    rcvStr.toLowerCase();
    rcvStr.toCharArray(rcvBuffer,255);
    sendCommand(rcvBuffer,14000);
    cmdClr();
  }
  else if (rcvStr.substring(0,4) == "BACK")  { //20 - 500 cm
    rcvStr.toLowerCase();
    rcvStr.toCharArray(rcvBuffer,255);
    sendCommand(rcvBuffer,14000);
    cmdClr();
  }
  else if (rcvStr.substring(0,2) == "UP")  {    //20 - 500 cm
    rcvStr.toLowerCase();
    rcvStr.toCharArray(rcvBuffer,255);
    sendCommand(rcvBuffer,14000);
    cmdClr();
  }
  else if (rcvStr.substring(0,4) == "DOWN")  {  //20 - 500 cm
    rcvStr.toLowerCase();
    rcvStr.toCharArray(rcvBuffer,255);
    sendCommand(rcvBuffer,14000);
    cmdClr();
  }
else if (rcvStr.substring(0,4) == "LEFT")  {   //20 - 500 cm
    rcvStr.toLowerCase();
    rcvStr.toCharArray(rcvBuffer,255);
    sendCommand(rcvBuffer,14000);
    cmdClr();
  }
  else if (rcvStr.substring(0,5) == "RIGHT")  { //20 - 500 cm
    rcvStr.toLowerCase();
    rcvStr.toCharArray(rcvBuffer,255);
    sendCommand(rcvBuffer,14000);
    cmdClr();
  }
  else if (rcvStr.substring(0,4) == "FLIP")  { //L / R / F / B
    rcvStr.toLowerCase();
    rcvStr.toCharArray(rcvBuffer,255);
    sendCommand(rcvBuffer,14000);
    cmdClr();
  }
  else if (rcvStr.substring(0,2) == "GO")  { //x y z speed
    rcvStr.toLowerCase();
    rcvStr.toCharArray(rcvBuffer,255);
    sendCommand(rcvBuffer,14000);
    cmdClr();
  }
  else if (rcvStr.substring(0,5) == "CURVE")  { //x1 y1 z1 x2 y2 z2 speed
    rcvStr.toLowerCase();
    rcvStr.toCharArray(rcvBuffer,255);
    sendCommand(rcvBuffer,14000);
    cmdClr();
  }
  else if (rcvStr.substring(0,5) == "SPEED")  { // speed 10..100cm/s
    rcvStr.toLowerCase();
    rcvStr.toCharArray(rcvBuffer,255);
    sendCommand(rcvBuffer,14000);
    cmdClr();
  }
/************************************************  
 *          Tello EDU SDK 2.0 control commands  *
 ************************************************/
  else if (rcvStr.substring(0,3) == "MON")  { // mission pad detect on
    rcvStr.toLowerCase();
    rcvStr.toCharArray(rcvBuffer,255);
    sendCommand(rcvBuffer,14000);
    cmdClr();
  }
  else if (rcvStr.substring(0,4) == "MOFF")  { // mission pad detect off
    rcvStr.toLowerCase();
    rcvStr.toCharArray(rcvBuffer,255);
    sendCommand(rcvBuffer,14000);
    cmdClr();
  }
  else if (rcvStr.substring(0,10) == "MDIRECTION")  { // camera select (0/1/2)
    rcvStr.toLowerCase();
    rcvStr.toCharArray(rcvBuffer,255);
    sendCommand(rcvBuffer,14000);
    cmdClr();
  }
  else if (rcvStr.substring(0,4) == "JUMP")  { // camera select (0/1/2)
    rcvStr.toLowerCase();
    rcvStr.toCharArray(rcvBuffer,255);
    sendCommand(rcvBuffer,14000);
    cmdClr();
  }
  
  // Read status commands
  else if (rcvStr == "BAT?")  {
    sendMessage("battery?");
    Tmessage = listenMessage(3000);
    if (Tmessage=="ok") Tmessage = listenMessage(3000);
    Serial.print("Batt level %: "+Tmessage);  
    cmdClr();
  }
  else if (rcvStr == "TEMP?")  {
    sendMessage("temp?");
    Tmessage = listenMessage(3000);
    if (Tmessage=="ok") Tmessage = listenMessage(3000);
    Serial.print("TEMPERATURE: "+Tmessage);
    cmdClr();
  }
  else if (rcvStr == "SPEED?")  {
    sendMessage("speed?");
    Tmessage = listenMessage(3000);
    if (Tmessage=="ok") Tmessage = listenMessage(3000);
    Serial.print(Tmessage);
    cmdClr();
  }
  else if (rcvStr == "TIME?")  {
    sendMessage("time?");
    Tmessage = listenMessage(3000);
    if (Tmessage=="ok") Tmessage = listenMessage(3000);
    Serial.print("FLIGHT TIME: "+Tmessage);
    cmdClr();
  }
  else if (rcvStr == "HEIGHT?")  {
    sendMessage("height?");
    Tmessage = listenMessage(3000);
    if (Tmessage=="ok") Tmessage = listenMessage(3000);
    Serial.print("HEIGHT: "+Tmessage);
    cmdClr();
  }
  else if (rcvStr == "ATT?")  {
    sendMessage("attitude?");
    Tmessage = listenMessage(3000);
    if (Tmessage=="ok") Tmessage = listenMessage(3000);
    Tmessage.toUpperCase();
    Serial.print(Tmessage);
    cmdClr();
  }
  else if (rcvStr == "BARO?")  {
    sendMessage("baro?");
    Tmessage = listenMessage(3000);
    if (Tmessage=="ok") Tmessage = listenMessage(3000);
    Serial.print("BAROMETER: "+Tmessage);
    cmdClr();
  }
  else if (rcvStr == "ACC?")  {
    sendMessage("acceleration?");
    Tmessage = listenMessage(3000);
    if (Tmessage=="ok") Tmessage = listenMessage(3000);
    Tmessage.toUpperCase();
    Serial.print(Tmessage);
    cmdClr();
  }
  else if (rcvStr == "TXF?")  {  //for unkown reasons, UART does not like tof
    sendMessage("tof?");
    Tmessage = listenMessage(3000);
    if (Tmessage=="ok") Tmessage = listenMessage(3000);
    Serial.print("DIST FROM TOF: "+Tmessage);
    cmdClr();
  }
  else if (rcvStr == "STATE?")  {
    if (TelloOnline(1))  {
      Serial.print(listenMessager(2000));
    }
    cmdClr();
  }
  /* ********************************************  
 *          Tello EDU SDK 2.0 read commands     *
 ************************************************/
  else if (rcvStr == "SDK?")  {
    sendMessage("sdk?");
    Tmessage = listenMessage(3000);
    if (Tmessage=="ok") Tmessage = listenMessage(3000);
    Serial.print("SDK VERS: "+Tmessage);
    cmdClr();
  }
  
  
  LEDstatus(LEDcommand);
}

// Wifi functions
bool initSequence() {
  if (connectWiFi())  {
    Udp.begin(PORT);
    Udpr.begin(PORTr);
    if (TelloOnline(5)) {  // five attempts to flush Tello onboard SDK
      return true;
    }
  }
  return false;
}

bool connectWiFi() {
  WiFi.begin(DEFAULT_SSID, DEFAULT_PW);
  int wifiTag = millis();
  while ((WiFi.status() != WL_CONNECTED) and (millis()<wifiTag+wifiTimeout))  {
    delay(500);
    if (debug)  {
      Serial.println("WIFI?");
    }
  }
  if (millis()>wifiTag+wifiTimeout)  {
    return false;
  }
  else  {
  if (debug) {
    Serial.println("CONNECTED TO WIFI");
    //delay(100);
    Serial.print("STATUS: "); Serial.println(WiFi.status());
    //delay(100);
    Serial.print("IP: ");     Serial.println(WiFi.localIP());
    //delay(100);
    Serial.print("SUBNET: "); Serial.println(WiFi.subnetMask());
    //delay(100);
    Serial.print("GATEWAY: "); Serial.println(WiFi.gatewayIP());
    //delay(100);
    Serial.print("SSID: "); Serial.println(WiFi.SSID());
    //delay(100);
    Serial.printf("SIGNAL: %d\n",WiFi.RSSI());
    //delay(100);
  }
  networkState = WIFIOK;
  wififresh = false;
  delay(1000);
  return true;
  }
}

// UDP read & write packet functions
String listenMessage(int msgTimeout) {
  int msgTag = millis();
  int packetSize = Udp.parsePacket();
  while ((millis()<(msgTag+msgTimeout)) and (packetSize==0))  {
    packetSize = Udp.parsePacket();
    if (debug)  {
      Serial.println("W");
    }
    LEDstatus(LEDUDP);
    delay(100);
  }
  if (packetSize) {
    // read the packet into packetBufffer
    int len = Udp.read(packetBuffer, 255);
    if (len > 0) {
      packetBuffer[len] = '\0';
    } 
    if (debug)  {
      Serial.print("CONTENTS:");
      Serial.println(packetBuffer);
    }
  }
  else  {
    packetBuffer[0] = '\0';
  }
  return (char*) packetBuffer;
}

String listenMessager(int msgTimeout) {
  int msgTag = millis();
  int packetSize = Udpr.parsePacket();
  while ((millis()<(msgTag+msgTimeout)) and (packetSize==0))  {
    packetSize = Udp.parsePacket();
    if (debug)  {
      Serial.println("WR");
    }
    LEDstatus(LEDUDP);
    delay(100);
  }
  if (packetSize) {
    // read the packet into packetBufffer
    int len = Udpr.read(packetBuffer, 255);
    if (len > 0) {
      packetBuffer[len] = '\0';
    }
    if (debug)  {
      Serial.print("R CONTENTS:");
      Serial.println(packetBuffer);
    } 
  }
  else  {
    packetBuffer[0] = '\0';
  }
  return (char*) packetBuffer;
}

void sendMessage(char* ReplyBuffer) {
  if (debug)  {
    Tmessage = ReplyBuffer;
    Serial.println("cmd: "+Tmessage);
  }
  Udp.beginPacket(TELLO_IP, PORT);
  Udp.write(ReplyBuffer);
  Udp.endPacket();
}

// CLI management functions
bool TelloOnline(int attempts)  {
  for (int i = 0; i < attempts; i++)  {
    sendMessage("command");
    if (listenMessage(1500)=="ok")  {
      networkState = CONNECTED;
   // i = attempts;
      return true;
    }
  }
  networkState = LOST;
  return false; 
}

void sendCommand(char* cmd,int cmdexe)  {
  sendMessage(cmd);
  Serial.print(listenMessage(cmdexe));
}

void cmdClr()  {
  rcvStr = "";
  Ctag = millis();
}

void LEDstatus (int Ltimeout)  {
  digitalWrite(ONBOARD_LED,flip);
  if (millis()>LEDtag+Ltimeout)  {
    flip=!flip;
    LEDtag = millis();
  }
}
